package com.blockchain.server.btc.service.impl;

import com.blockchain.common.base.dto.GasDTO;
import com.blockchain.common.base.dto.WalletChangeDTO;
import com.blockchain.common.base.dto.WalletOrderDTO;
import com.blockchain.server.btc.common.constants.BtcTransferConstans;
import com.blockchain.server.btc.common.enums.BtcEnums;
import com.blockchain.server.btc.common.exception.BtcException;
import com.blockchain.server.btc.common.util.CheckEthFeginResult;
import com.blockchain.server.btc.dto.BtcWalletDTO;
import com.blockchain.server.btc.dto.BtcWalletTransferDTO;
import com.blockchain.server.btc.entity.BtcWalletTransfer;
import com.blockchain.server.btc.feign.EthServerFegin;
import com.blockchain.server.btc.mapper.BtcWalletTransferMapper;
import com.blockchain.server.btc.service.*;
import com.codingapi.tx.annotation.ITxTransaction;
import com.codingapi.tx.annotation.TxTransaction;
import com.github.pagehelper.PageHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.UUID;

@Service
public class BtcWalletTransferServiceImpl implements BtcWalletTransferService, ITxTransaction {
    @Autowired
    private BtcWalletTransferMapper btcWalletTransferMapper;

    @Autowired
    private BtcWalletService btcWalletService;

    @Autowired
    private BtcApplicationService btcApplicationService;

    @Autowired
    private BtcTokenService btcTokenService;

    @Autowired
    private ConfigWalletParamService walletParamService;

    @Autowired
    private EthServerFegin ethServerFegin;

    @Override
    public Integer insertTransfer(BtcWalletTransfer btcWalletTransfer) {
        return btcWalletTransferMapper.insertSelective(btcWalletTransfer);
    }

    @Override
    public List<BtcWalletTransferDTO> selectTransfer(String userOpenId, Integer tokenId, String walletType) {
        //校验钱包类型
        btcApplicationService.checkWalletType(walletType);
        //校验币种id
        btcTokenService.getAndVerifyTokenNameById(tokenId);
        BtcWalletDTO btcWalletDTO = btcWalletService.selectByUserOpenId(userOpenId, tokenId, walletType);
        String addr = btcWalletDTO.getAddr();
        return btcWalletTransferMapper.selectTransfer(addr, tokenId);
    }

    @Override
    public Integer updateStatusById(String id, int status) {
        BtcWalletTransfer btcWalletTransfer = new BtcWalletTransfer();
        btcWalletTransfer.setId(id);
        btcWalletTransfer.setStatus(status);
        return btcWalletTransferMapper.updateByPrimaryKeySelective(btcWalletTransfer);
    }

    @Override
    public Boolean isNotExistsTransferIn(String txId) {
        String id = btcWalletTransferMapper.selectIdByTxIdAndType(txId, BtcTransferConstans.TYPE_IN);
        return StringUtils.isEmpty(id);
    }

    @Override
    @Transactional
    @TxTransaction
    public BtcWalletDTO handleOrder(WalletOrderDTO walletOrderDTO) {
        double freeBalance = walletOrderDTO.getFreeBalance().doubleValue();
        double freezeBalance = walletOrderDTO.getFreezeBalance().doubleValue();

        int tokenId = btcTokenService.getAndVerifyTokenIdByName(walletOrderDTO.getTokenName());
        //查询用户钱包
        BtcWalletDTO btcWalletDTO = btcWalletService.selectByUserOpenId(walletOrderDTO.getUserId(), tokenId, walletOrderDTO.getWalletType());
        if (btcWalletDTO == null) {
            throw new BtcException(BtcEnums.NO_WALLET_ERROR);
        }

        if (btcWalletDTO.getFreeBalance() + freeBalance < 0 || btcWalletDTO.getFreezeBalance() + freezeBalance < 0) {
            throw new BtcException(BtcEnums.FREEBALANCE_NOT_ENOUGH);
        }

        //加减钱包可用余额、冻结余额，之间转换
        int countUb = btcWalletService.updateBalanceByAddrInRowLock(btcWalletDTO.getAddr(), btcWalletDTO.getTokenId(), freeBalance, freezeBalance, 0.0, new Date());
        if (countUb != 1) {
            throw new BtcException(BtcEnums.TRANSFER_ERROR);
        }

        //返回加减余额后的数据
        return btcWalletService.selectByUserOpenId(walletOrderDTO.getUserId(), tokenId, walletOrderDTO.getWalletType());
    }

    @Override
    @Transactional
    @TxTransaction
    public Integer handleChange(WalletChangeDTO walletChangeDTO) {
        double freeBalance = walletChangeDTO.getFreeBalance().doubleValue();
        double freezeBalance = walletChangeDTO.getFreezeBalance().doubleValue();
        double totalBalance = freeBalance + freezeBalance;

        Date now = new Date();

        int tokenId = btcTokenService.getAndVerifyTokenIdByName(walletChangeDTO.getTokenName());
        //查询用户钱包
        BtcWalletDTO btcWalletDTO = btcWalletService.selectByUserOpenId(walletChangeDTO.getUserId(), tokenId, walletChangeDTO.getWalletType());
        if (btcWalletDTO == null) {
            throw new BtcException(BtcEnums.NO_WALLET_ERROR);
        }

        if (btcWalletDTO.getFreeBalance() + freeBalance < 0 || btcWalletDTO.getFreezeBalance() + freezeBalance < 0) {
            throw new BtcException(BtcEnums.BALANCE_NOT_ENOUGH);
        }

        if (freeBalance != 0 || freezeBalance != 0) {
            //加减钱包可用余额、冻结余额、总额
            int countUb = btcWalletService.updateBalanceByAddrInRowLock(btcWalletDTO.getAddr(), btcWalletDTO.getTokenId(), freeBalance, freezeBalance, totalBalance, now);
            if (countUb != 1) {
                throw new BtcException(BtcEnums.TRANSFER_ERROR);
            }
        }

        //插入一条交易记录
        BtcWalletTransfer btcWalletTransfer = new BtcWalletTransfer();
        if (totalBalance < 0) {
            btcWalletTransfer.setFromAddr(btcWalletDTO.getAddr());
            btcWalletTransfer.setToAddr(null);
        } else {
            btcWalletTransfer.setFromAddr(null);
            btcWalletTransfer.setToAddr(btcWalletDTO.getAddr());
        }
        btcWalletTransfer.setId(UUID.randomUUID().toString());
        btcWalletTransfer.setHash(walletChangeDTO.getRecordId());

        btcWalletTransfer.setAmount(Math.abs(totalBalance));
        btcWalletTransfer.setTokenId(tokenId);
        btcWalletTransfer.setTokenSymbol(walletChangeDTO.getTokenName());
        btcWalletTransfer.setGasPrice(walletChangeDTO.getGasBalance().doubleValue());
        btcWalletTransfer.setTransferType(walletChangeDTO.getWalletType());
        btcWalletTransfer.setStatus(BtcTransferConstans.STATUS_SUCCESS);
        btcWalletTransfer.setCreateTime(now);
        btcWalletTransfer.setUpdateTime(now);
        int countIt = insertTransfer(btcWalletTransfer);

        if (countIt != 1) {
            throw new BtcException(BtcEnums.TRANSFER_ERROR);
        }

        return 1;
    }

    @Override
    @Transactional
    public BtcWalletDTO handleWithdraw(String userOpenId, String password, String toAddress, Integer tokenId, double amount, String walletType) {
        //校验钱包类型
        btcApplicationService.checkWalletType(walletType);
        //校验币种id
        String tokenSymbol = btcTokenService.getAndVerifyTokenNameById(tokenId);

        //获取提现手续费
        GasDTO gasDTO = walletParamService.getGasConfig(tokenSymbol);
        //与最小提现数额做对比
        if (gasDTO.getMinWdAmount().compareTo(new BigDecimal(amount)) > 0) {
            throw new BtcException(BtcEnums.LOW_WITHDRAW_AMOUNT_ERROR);
        }

        //检验密码
        CheckEthFeginResult.checkIsPassword(ethServerFegin.isPassword(password));

        //验证提现地址是否有效
//        JSONObject vaObj = null;
//        try {
//            vaObj = btcUtils.validateAddress(toAddress);
//        } catch (Exception e) {
//            throw new BtcException(BtcEnums.ADDRESS_ERROR);
//        }
//        //验证提现地址是否有效
//        if (!vaObj.getBoolean("isvalid")) {
//            throw new BtcException(BtcEnums.ADDRESS_ERROR);
//        }

        //获取钱包余额
        BtcWalletDTO btcWalletDTO = btcWalletService.selectByUserOpenId(userOpenId, tokenId, walletType);
        if (btcWalletDTO.getFreeBalance() + amount < 0) {
            throw new BtcException(BtcEnums.FREEBALANCE_NOT_ENOUGH);
        }
        //用户转出钱包地址
        String fromAddress = btcWalletDTO.getAddr();

        Date now = new Date();

        //*********** 是，快速转账，数据库划转 ***********

        //该用户减去提现可用余额、总额
        int countUb = btcWalletService.updateBalanceByAddrInRowLock(fromAddress, tokenId, -amount, 0.0, -amount, now);
        if (countUb != 1) {
            throw new BtcException(BtcEnums.WITHDRAW_ERROR);
        }
//        //接收用户加上可用余额、总额
//        int countUbR = btcWalletService.updateBalanceByAddrInRowLock(toAddress, tokenId, amount, 0.0, amount, now);
//        if (countUbR != 1) {
//            throw new BtcException(BtcEnums.WITHDRAW_ERROR);
//        }
        //不走链上,快速到账。
        //并插入一条提现记录，站内快速转账
        BtcWalletTransfer btcWalletTransfer = new BtcWalletTransfer();
        btcWalletTransfer.setId(UUID.randomUUID().toString());
//            btcWalletTransfer.setHash(null);
        btcWalletTransfer.setFromAddr(fromAddress);
        btcWalletTransfer.setToAddr(toAddress);
        btcWalletTransfer.setAmount(Math.abs(amount));
//            btcWalletTransfer.setGasPrice(null);
        btcWalletTransfer.setTokenId(tokenId);
        btcWalletTransfer.setTokenSymbol(tokenSymbol);
        btcWalletTransfer.setTransferType(BtcTransferConstans.TYPE_FAST);
        btcWalletTransfer.setStatus(BtcTransferConstans.STATUS_SUCCESS);
        btcWalletTransfer.setCreateTime(now);
        btcWalletTransfer.setUpdateTime(now);
        int countIt = insertTransfer(btcWalletTransfer);
        if (countIt != 1) {
            throw new BtcException(BtcEnums.WITHDRAW_ERROR);
        }
        //判断是否属于节点钱包
//        if (vaObj.getBoolean("ismine")) {
//        } else {
//            //*********** 否，区块链转账 ***********
//            // 查询是否存在提现黑名单中
//            // 抛出错误表示用户禁止提现
//            userServerFegin.verifyBanWithdraw(userOpenId);
//            // 判断用户是否存在提现白名单中
//            ResultDTO<Boolean> resultDTO = userServerFegin.verifyFreeWithdraw(userOpenId);
//            if (resultDTO.getCode() != BaseConstant.REQUEST_SUCCESS)
//                throw new RPCException(resultDTO.getCode(), resultDTO.getMsg());
//            // 用户存在提现白名单中，设置提现手续费为零
//            if (resultDTO.getData()) gasDTO.setGasPrice(BigDecimal.ZERO);
//            // 余额加上扣除手续费
//            //该用户减去提现可用余额、加上冻结余额
//            int countUb = btcWalletService.updateBalanceByAddrInRowLock(fromAddress, tokenId, -amount, amount, 0.0, now);
//            if (countUb != 1) {
//                throw new BtcException(BtcEnums.WITHDRAW_ERROR);
//            }
//            //并插入一条提现记录
//            BtcWalletTransfer btcWalletTransfer = new BtcWalletTransfer();
//            btcWalletTransfer.setId(UUID.randomUUID().toString());
////            btcWalletTransfer.setHash(null);
//            btcWalletTransfer.setFromAddr(fromAddress);
//            btcWalletTransfer.setToAddr(toAddress);
//            btcWalletTransfer.setAmount(Math.abs(amount));
//            btcWalletTransfer.setTokenId(tokenId);
//            btcWalletTransfer.setTokenSymbol(tokenSymbol);
//            //设置手续费
//            //TODO 扣减不同币种手续费
//            btcWalletTransfer.setGasPrice(gasDTO.getGasPrice().doubleValue());
//            btcWalletTransfer.setGasTokenType(gasDTO.getGasTokenType());
//            btcWalletTransfer.setGasTokenSymbol(gasDTO.getGasTokenSymbol());
//            btcWalletTransfer.setGasTokenName(gasDTO.getGasTokenName());
//            btcWalletTransfer.setTransferType(BtcTransferConstans.TYPE_OUT);
//            btcWalletTransfer.setStatus(BtcTransferConstans.STATUS_FIRST_TRIAL);
//            btcWalletTransfer.setCreateTime(now);
//            btcWalletTransfer.setUpdateTime(now);
//            int countIt = insertTransfer(btcWalletTransfer);
//            if (countIt != 1) {
//                throw new BtcException(BtcEnums.WITHDRAW_ERROR);
//            }
//        }

        //返回加减余额后的数据
        return btcWalletService.selectByAddr(fromAddress, tokenId, walletType);
    }

    /*@Override
    public List<BtcWalletTransfer> selectByTxTypeAndStatus(String transferType, int Status, int pageNum, int pageSize) {
        PageHelper.startPage(pageNum, pageSize);  // 分页处理
        BtcWalletTransfer where = new BtcWalletTransfer();
        where.setTransferType(transferType);
        where.setStatus(Status);
        return btcWalletTransferMapper.select(where);
    }*/

    @Override
    @Transactional
    public void updateStatus(String id, int status, Date date) {
        BtcWalletTransfer ethWalletTransfer = btcWalletTransferMapper.selectByPrimaryKey(id);
        ethWalletTransfer.setStatus(status);
        ethWalletTransfer.setUpdateTime(date);
        // 失败抛出异常
        int row = btcWalletTransferMapper.updateByPrimaryKeySelective(ethWalletTransfer);
        if (row == 0) {
            throw new BtcException(BtcEnums.SERVER_IS_TOO_BUSY);
        }
    }

}
