package com.blockchain.server.cct.service.impl;

import com.blockchain.common.base.dto.ResultDTO;
import com.blockchain.server.cct.common.enums.CctDataEnums;
import com.blockchain.server.cct.common.enums.CctEnums;
import com.blockchain.server.cct.common.exception.CctException;
import com.blockchain.server.cct.dto.rpc.currency.CurrencyMarketDTO;
import com.blockchain.server.cct.entity.Commission;
import com.blockchain.server.cct.entity.rpc.user.UserRelation;
import com.blockchain.server.cct.feign.CurrencyFeign;
import com.blockchain.server.cct.feign.UserFeign;
import com.blockchain.server.cct.mapper.CommissionMapper;
import com.blockchain.server.cct.service.CommissionService;
import com.blockchain.server.cct.service.ConfigService;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.UUID;

@Service
public class CommissionServiceImpl implements CommissionService {

    @Autowired
    private CommissionMapper commissionMapper;
    @Autowired
    private ConfigService configService;
    @Autowired
    private UserFeign userFeign;
    @Autowired
    private CurrencyFeign currencyFeign;

    @Override
    @Transactional
    public void generateCommission(String userId, String recordId, BigDecimal serviceCharge, String coinName) {
        //判断用户id是否为空
        checkUserIdNull(userId);
        //判断代币是否为空
        checkCoinNameNull(coinName);
        //判断手续费是否无法生成佣金
        if (!checkServiceCharge(serviceCharge)) {
            return;
        }

        //用户关系链
        UserRelation userRelation = selectRelationByUserId(userId);
        //关系链不存在或者父级id不存在
        if (userRelation == null || StringUtils.isBlank(userRelation.getPid())) {
            return;
        }

        //查询用户注册时间是否超过佣金有效期
        boolean validity = checkCommissionValidity(userRelation.getCreateTime());
        if (!validity) {
            return;
        }

        //查询佣金代币
        String commissionCoin = configService.selectCommissionCoin();
        if (StringUtils.isBlank(commissionCoin)) {
            return;
        }
        //查询佣金百分比
        BigDecimal commissionCharge = configService.selectCommissionCharge();
        if (commissionCharge == null) {
            return;
        }

        //计算佣金
        BigDecimal commissionAmount = computeCommissionAmount(serviceCharge, coinName, commissionCharge, commissionCoin);

        //判断是否存在
        Commission commission = new Commission();
        Date now = new Date();
        commission.setId(UUID.randomUUID().toString());
        commission.setUserId(userId);
        commission.setPid(userRelation.getPid());
        commission.setRecordId(recordId);
        commission.setAmount(commissionAmount);
        commission.setCoinName(commissionCoin);
        commission.setCreateTime(now);
        commission.setModifyTime(now);
        commission.setStatus(CctDataEnums.COMMON_STATUS_NO.getStrVlue());
        commissionMapper.insertSelective(commission);
    }

    /***
     * 判断用户id是否为空
     * @param userId
     */
    private void checkUserIdNull(String userId) {
        if (StringUtils.isBlank(userId)) {
            throw new CctException(CctEnums.ORDER_USERID_NULL);
        }
    }

    /***
     * 判断币种是否为空
     * @param coinName
     */
    private void checkCoinNameNull(String coinName) {
        if (StringUtils.isBlank(coinName)) {
            throw new CctException(CctEnums.ORDER_COINNAME_NULL);
        }
    }

    /***
     * //判断手续费为空或者小于0
     * 返回false终止生成佣金方法
     * @param serviceCharge
     * @return
     */
    private boolean checkServiceCharge(BigDecimal serviceCharge) {
        if (serviceCharge == null || serviceCharge.compareTo(BigDecimal.ZERO) <= 0) {
            return false;
        }
        return true;
    }

    /***
     * 计算佣金
     * @param serviceCharge
     * @param coinName
     * @param commissionCharge
     * @param commissionCoin
     * @return
     */
    private BigDecimal computeCommissionAmount(BigDecimal serviceCharge, String coinName, BigDecimal commissionCharge, String commissionCoin) {
        //如果代币跟数据库设置的代币不一致，查询最新行情获取汇率
        if (!coinName.equals(commissionCoin)) {
            CurrencyMarketDTO currency = getCurrency(coinName, commissionCoin);
            //获取 成交代币/佣金代币 最新行情
            BigDecimal amount = currency.getAmount();
            //将成交手续费兑换成佣金代币 ： 成交手续费 * 最新行情
            serviceCharge = serviceCharge.multiply(amount);
        }

        //佣金 = 成交手续费 * 佣金百分比
        return serviceCharge.multiply(commissionCharge);
    }

    /***
     * 计算生成佣金用户注册时间是否超过有效期
     * @param registerTime
     * @return 返回false 不生成佣金
     */
    private boolean checkCommissionValidity(Date registerTime) {
        //查询佣金有效期
        Integer validity = configService.selectCommissionValidity();
        try {
            //计算用户注册至今的天数
            Integer difference = checkUserRegisterTimeDifference(registerTime);
            //如果用户注册至今的天数超过佣金生成有效期
            if (difference > validity) {
                return false;
            }
            return true;
        } catch (ParseException e) {
            return false;
        }
    }

    /***
     * 计算用户注册至今的时间天数
     * @param registerTime
     * @return
     * @throws ParseException
     */
    private Integer checkUserRegisterTimeDifference(Date registerTime) throws ParseException {
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
        //通过格式化去除时分秒
        String registerTimeFormat = dateFormat.format(registerTime);
        String nowTimeFormat = dateFormat.format(new Date());
        //再转回日期
        registerTime = dateFormat.parse(registerTimeFormat);
        Date nowTime = dateFormat.parse(nowTimeFormat);
        //差再除以1天的毫秒数
        return (int) ((nowTime.getTime() - registerTime.getTime()) / (1000 * 3600 * 24));
    }

    /***
     * 查询最新行情
     * @param coinName
     * @param unitName
     * @return
     */
    private CurrencyMarketDTO getCurrency(String coinName, String unitName) {
        String currencyPair = "{0}-{1}";
        ResultDTO<CurrencyMarketDTO> resultDTO = currencyFeign.get(MessageFormat.format(currencyPair, coinName, unitName));
        return resultDTO.getData();
    }

    /***
     * 根据用户查询推荐关系
     * @param userId
     * @return
     */
    private UserRelation selectRelationByUserId(String userId) {
        ResultDTO<UserRelation> resultDTO = userFeign.selectRelationByUserId(userId);
        return resultDTO.getData();
    }
}
