package com.blockchain.server.yyyf.indicators.handle;

import com.blockchain.server.yyyf.indicators.IndicatorsHandler;
import com.blockchain.server.yyyf.indicators.dto.AssetsReporter;
import com.blockchain.server.yyyf.service.AssetsReporterService;
import lombok.AllArgsConstructor;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.stereotype.Component;

import javax.persistence.EntityExistsException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/**
 * @author fengyu
 */
@AllArgsConstructor
@Component(FluctuationRateIndicatorsHandler.HANDLER_NAME + IndicatorsHandler.HANDLER_NAME)
public class FluctuationRateIndicatorsHandler implements IndicatorsHandler {
    public static final String HANDLER_NAME = "FluctuationRate";
    public static final String CACHE_FLUCTUATION_RATE = "yyyf_indicator_fluctuationRate";
    public static final String CACHE_KEY_FLUCTUATION_RATE = "yyyf:fluctuationRate:";

    private final AssetsReporterService assetsReporterService;

    @Override
    @Cacheable(cacheNames = CACHE_FLUCTUATION_RATE,
            key = "#root.target.CACHE_KEY_FLUCTUATION_RATE.concat(#assessUserId)", unless = "#result == null")
    public Float handleIndicator(String assessUserId) {
        List<AssetsReporter> reporterList = assetsReporterService.selectAssetsReporterListByAssessUserId(assessUserId);
        if (reporterList != null && reporterList.size() > 0) {
            //排序
            reporterList.sort(Comparator.comparingLong(AssetsReporter::getDays));
            //总天数
            int dayCount = reporterList.size() - 1;
            //每日收益率List
            List<BigDecimal> dayYieldList = new ArrayList<BigDecimal>(dayCount);
            //收益率总和
            BigDecimal totalYield = BigDecimal.ZERO;
            //上一天的资产总额
            BigDecimal upAssets = reporterList.get(0).getAssets();

            //计算总收益率以及每日收益率
            for (AssetsReporter assetsReporter : reporterList) {
                if (assetsReporter.getDays() > 0) {
                    BigDecimal thisAssets = assetsReporter.getAssets();
                    BigDecimal subtract = thisAssets.subtract(upAssets);
                    //当天收益率
                    BigDecimal thisYeid = subtract.multiply(BigDecimal.valueOf(100)).divide(upAssets, 5, BigDecimal.ROUND_HALF_UP);
                    //加入每日收益率集合
                    dayYieldList.add(thisYeid);
                    //总收益率
                    totalYield = totalYield.add(thisYeid);
                    //更换上一天的总额
                    upAssets = thisAssets;
                }
            }
            //平均收益率
            BigDecimal avgYeid = totalYield.divide(BigDecimal.valueOf(dayCount), 5, BigDecimal.ROUND_HALF_UP);
            //（每日收益率-平均收益率)的平方和
            BigDecimal yeid2Count = BigDecimal.ZERO;
            for (BigDecimal yeid : dayYieldList) {
                yeid2Count = yeid2Count.add(yeid.subtract(avgYeid).pow(2));
            }
            return Float.valueOf(String.format("%.2f", Math.sqrt(BigDecimal.valueOf(360).divide(BigDecimal.valueOf(dayCount - 1), 5, BigDecimal.ROUND_HALF_UP).multiply(yeid2Count).doubleValue())));
        } else {
            throw new EntityExistsException("assessUser:" + assessUserId);
        }
    }

}
